# coding: utf-8
import base64
from enum import Enum
import zlib
import io
import cv2
import numpy as np
from distutils.version import StrictVersion

from PIL import Image
from skimage import morphology as skimage_morphology

from supervisely_lib.geometry.bitmap_base import BitmapBase, resize_origin_and_bitmap
from supervisely_lib.geometry.point_location import PointLocation, row_col_list_to_points
from supervisely_lib.geometry.polygon import Polygon
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.constants import BITMAP


class SkeletonizeMethod(Enum):
    SKELETONIZE = 0
    MEDIAL_AXIS = 1
    THINNING = 2


def _find_mask_tight_bbox(raw_mask: np.ndarray) -> Rectangle:
    rows = list(np.any(raw_mask, axis=1).tolist())  # Redundant conversion to list to help PyCharm static analysis.
    cols = list(np.any(raw_mask, axis=0).tolist())
    top_margin = rows.index(True)
    bottom_margin = rows[::-1].index(True)
    left_margin = cols.index(True)
    right_margin = cols[::-1].index(True)
    return Rectangle(top=top_margin, left=left_margin, bottom=len(rows) - 1 - bottom_margin,
                     right=len(cols) - 1 - right_margin)


class Bitmap(BitmapBase):
    '''
    This is a class for creating and using Bitmaps for Labels.
    '''
    @staticmethod
    def geometry_name():
        return 'bitmap'

    def __init__(self, data: np.ndarray, origin: PointLocation = None,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        The constructor for Bitmap class.
        :param data: bool numpy array
        :param origin: points (x and y coordinates) of the top left corner of a bitmap, i.e. the position of the
        bitmap within the image
        '''

        if data.dtype != np.bool:
            raise ValueError('Bitmap mask data must be a boolean numpy array. Instead got {}.'.format(str(data.dtype)))

        # Call base constructor first to do the basic dimensionality checks.
        super().__init__(data=data, origin=origin, expected_data_dims=2,
                         sly_id=sly_id, class_id=class_id,
                         labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        # Crop the empty margins of the boolean mask right away.
        if not np.any(data):
            raise ValueError('Creating a bitmap with an empty mask (no pixels set to True) is not supported.')
        data_tight_bbox = _find_mask_tight_bbox(self._data)
        self._origin = self._origin.translate(drow=data_tight_bbox.top, dcol=data_tight_bbox.left)
        self._data = data_tight_bbox.get_cropped_numpy_slice(self._data)

    @classmethod
    def _impl_json_class_name(cls):
        return BITMAP

    def rotate(self, rotator):
        '''
        The function rotate Bitmap mask with a given rotator
        (ImageRotator class object contain size of image and angle to rotate)
        :param rotator: ImageRotator class object
        :return: Bitmap class object
        '''
        full_img_mask = np.zeros(rotator.src_imsize, dtype=np.uint8)
        self.draw(full_img_mask, 1)
        # TODO this may break for one-pixel masks (it can disappear during rotation). Instead, rotate every pixel
        #  individually and set it in the resulting bitmap.
        new_mask = rotator.rotate_img(full_img_mask, use_inter_nearest=True).astype(np.bool)
        return Bitmap(data=new_mask)

    def crop(self, rect):
        '''
        Crop the current Bitmap object with a given rectangle
        :param rect: Rectangle class object
        :return: Bitmap class object
        '''
        maybe_cropped_bbox = self.to_bbox().crop(rect)
        if len(maybe_cropped_bbox) == 0:
            return []
        else:
            [cropped_bbox] = maybe_cropped_bbox
            cropped_bbox_relative = cropped_bbox.translate(drow=-self.origin.row, dcol=-self.origin.col)
            cropped_mask = cropped_bbox_relative.get_cropped_numpy_slice(self._data)
            if not np.any(cropped_mask):
                return []
            return [Bitmap(data=cropped_mask, origin=PointLocation(row=cropped_bbox.top, col=cropped_bbox.left))]

    def resize(self, in_size, out_size):
        '''
        Resize the current Bitmap to match a certain size
        :param in_size: input image size
        :param out_size: output image size
        :return: Bitmap class object
        '''
        scaled_origin, scaled_data = resize_origin_and_bitmap(self._origin, self._data.astype(np.uint8), in_size,
                                                              out_size)
        # TODO this might break if a sparse mask is resized too thinly. Instead, resize every pixel individually and set
        #  it in the resulting bitmap.
        return Bitmap(data=scaled_data.astype(np.bool), origin=scaled_origin)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        self.to_bbox().get_cropped_numpy_slice(bitmap)[self.data] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        if StrictVersion(cv2.__version__) >= StrictVersion('4.0.0'):
            contours, _ = cv2.findContours(self.data.astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        else:
            _, contours, _ = cv2.findContours(self.data.astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        if contours is not None:
            for cont in contours:
                cont[:, :] += (self.origin.col, self.origin.row)  # cont with shape (rows, ?, 2)
            cv2.drawContours(bitmap, contours, -1, color, thickness=thickness)

    @property
    def area(self):
        '''
        :return: area of current Bitmap
        '''
        return float(self._data.sum())

    @staticmethod
    def base64_2_data(s: str) -> np.ndarray:
        '''
        The function base64_2_data convert base64 encoded string to numpy
        :param s: string
        :return: numpy array
        '''
        z = zlib.decompress(base64.b64decode(s))
        n = np.frombuffer(z, np.uint8)

        imdecoded = cv2.imdecode(n, cv2.IMREAD_UNCHANGED)
        if (len(imdecoded.shape) == 3) and (imdecoded.shape[2] >= 4):
            mask = imdecoded[:, :, 3].astype(bool)  # 4-channel imgs
        elif len(imdecoded.shape) == 2:
            mask = imdecoded.astype(bool)  # flat 2d mask
        else:
            raise RuntimeError('Wrong internal mask format.')
        return mask

    @staticmethod
    def data_2_base64(mask: np.ndarray) -> str:
        '''
        The function data_2_base64 convert numpy array to base64 encoded string
        :param mask: numpy array
        :return: string
        '''
        img_pil = Image.fromarray(np.array(mask, dtype=np.uint8))
        img_pil.putpalette([0, 0, 0, 255, 255, 255])
        bytes_io = io.BytesIO()
        img_pil.save(bytes_io, format='PNG', transparency=0, optimize=0)
        bytes_enc = bytes_io.getvalue()
        return base64.b64encode(zlib.compress(bytes_enc)).decode('utf-8')

    def skeletonize(self, method_id: SkeletonizeMethod):
        '''
        The function compute the skeleton, medial axis transform or morphological thinning of Bitmap
        :param method_id: method to convert binary image(numpy array)
        :return: Bitmap class object
        '''
        if method_id == SkeletonizeMethod.SKELETONIZE:
            method = skimage_morphology.skeletonize
        elif method_id == SkeletonizeMethod.MEDIAL_AXIS:
            method = skimage_morphology.medial_axis
        elif method_id == SkeletonizeMethod.THINNING:
            method = skimage_morphology.thin
        else:
            raise NotImplementedError('Method {!r} does\'t exist.'.format(method_id))

        mask_u8 = self.data.astype(np.uint8)
        res_mask = method(mask_u8).astype(bool)
        return Bitmap(data=res_mask, origin=self.origin)

    def to_contours(self):
        '''
        The function to_contours get list of contours(in Polygon class objects) of all objects in Bitmap
        :return: list of contours
        '''

        origin, mask = self.origin, self.data
        if StrictVersion(cv2.__version__) >= StrictVersion('4.0.0'):
            contours, hier = cv2.findContours(
                mask.astype(np.uint8),
                mode=cv2.RETR_CCOMP,  # two-level hierarchy, to get polygons with holes
                method=cv2.CHAIN_APPROX_SIMPLE
            )
        else:
            _, contours, hier = cv2.findContours(
                mask.astype(np.uint8),
                mode=cv2.RETR_CCOMP,  # two-level hierarchy, to get polygons with holes
                method=cv2.CHAIN_APPROX_SIMPLE
            )
        if (hier is None) or (contours is None):
            return []

        res = []
        for idx, hier_pos in enumerate(hier[0]):
            next_idx, prev_idx, child_idx, parent_idx = hier_pos
            if parent_idx < 0:
                external = contours[idx][:, 0].tolist()
                internals = []
                while child_idx >= 0:
                    internals.append(contours[child_idx][:, 0])
                    child_idx = hier[0][child_idx][0]
                if len(external) > 2:
                    new_poly = Polygon(exterior=row_col_list_to_points(external, flip_row_col_order=True),
                                       interior=[row_col_list_to_points(x, flip_row_col_order=True) for x in internals])
                    res.append(new_poly)

        offset_row, offset_col = origin.row, origin.col
        res = [x.translate(offset_row, offset_col) for x in res]
        return res

    def bitwise_mask(self, full_target_mask: np.ndarray, bit_op):
        '''
        The function bitwise_mask performs bitwise operations between a given numpy array and Bitmap object data
        :param full_target_mask: numpy array
        :param bit_op: type of bitwise operation(or, and, xor)
        :return: Bitmap class object or empty list
        '''
        full_size = full_target_mask.shape[:2]
        origin, mask = self.origin, self.data
        full_size_mask = np.full(full_size, False, bool)
        full_size_mask[origin.row:origin.row + mask.shape[0], origin.col:origin.col + mask.shape[1]] = mask

        new_mask = bit_op(full_target_mask, full_size_mask).astype(bool)
        if new_mask.sum() == 0:
            return []
        new_mask = new_mask[origin.row:origin.row + mask.shape[0], origin.col:origin.col + mask.shape[1]]
        return Bitmap(data=new_mask, origin=origin.clone())

    @classmethod
    def allowed_transforms(cls):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        from supervisely_lib.geometry.polygon import Polygon
        from supervisely_lib.geometry.rectangle import Rectangle
        return [AnyGeometry, Polygon, Rectangle]



from typing import Union
import supervisely_lib as sly
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.api.api import Api
from supervisely_lib.annotation.annotation import Annotation


class Gallery:
    def __init__(self, task_id, api: Api, v_model, project_meta: ProjectMeta, col_number: int, preview_info=False,
                 enable_zoom=False, resize_on_zoom=False,
                 sync_views=False, show_preview=True, selectable=False, opacity=0.5, show_opacity_header=True,
                 fillRectangle=False, borderWidth=3):
        self._task_id = task_id
        self._api = api
        self._v_model = v_model
        self._project_meta = project_meta.clone()
        self._data = {}
        self.col_number = col_number
        self.preview_info = preview_info
        self._need_zoom = False
        if not isinstance(self.col_number, int):
            raise ValueError("Columns number must be integer, not {}".format(type(self.col_number).__name__))

        self._options = {
            "enableZoom": enable_zoom,
            "syncViews": sync_views,
            "resizeOnZoom": resize_on_zoom,
            "showPreview": show_preview,
            "selectable": selectable,
            "opacity": opacity,
            "showOpacityInHeader": show_opacity_header,
            "fillRectangle": fillRectangle,
            "borderWidth": borderWidth
        }
        self._options_initialized = False

    def add_item(self, title, image_url, ann: Union[Annotation, dict] = None, col_index=None, info_dict=None,
                 zoom_to_figure=None):

        if col_index is not None:
            if col_index <= 0 or col_index > self.col_number:
                raise ValueError("Column number is not correct, check your input data")

        res_ann = Annotation((1, 1))
        if ann is not None:
            if type(ann) is dict:
                res_ann = Annotation.from_json(ann, self._project_meta)
            else:
                res_ann = ann.clone()

        self._data[title] = [image_url, res_ann, col_index]

        if zoom_to_figure is not None:
            self._data[title].append(zoom_to_figure)
            self._need_zoom = True

        if self.preview_info:
            if info_dict is not None:
                self._data[title].append(info_dict)

    def add_item_by_id(self, image_id, with_ann=True, col_index=None, info_dict=None,
                 zoom_to_figure=None):
        image_info = self._api.image.get_info_by_id(image_id)
        if with_ann:
            ann_info = self._api.annotation.download(image_id)
            ann = sly.Annotation.from_json(ann_info.annotation, self._project_meta)
        else:
            ann = None

        self.add_item(image_info.name, image_info.full_storage_url, ann, col_index, info_dict, zoom_to_figure)

    def _get_item_annotation(self, name):
        if self.preview_info:
            return {
                "url": self._data[name][0],
                "figures": [label.to_json() for label in self._data[name][1].labels],
                "title": name,
                "info": self._data[name][3]
            }
        else:
            return {
                "url": self._data[name][0],
                "figures": [label.to_json() for label in self._data[name][1].labels],
                "title": name,
            }

    def update(self, options=True):
        if len(self._data) == 0:
            raise ValueError("Items list is empty")

        gallery_json = self.to_json()

        if options is True or self._options_initialized is False:
            if self._need_zoom:
                self._options["resizeOnZoom"] = True
            self._api.task.set_field(self._task_id, self._v_model, gallery_json)
            self._options_initialized = True
        else:
            self._api.task.set_field(self._task_id, f"{self._v_model}.content", gallery_json["content"])

    def _zoom_to_figure(self, gallery_json):
        items = self._data.items()
        zoom_to_figure_name = "zoomToFigure"
        for item in items:
            curr_image_name = item[0]
            curr_image_data = item[1]

            if len(curr_image_data) < 4:
                raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))

            elif type(curr_image_data[3]) is not tuple:
                raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))

            zoom_params = {
                "figureId": curr_image_data[3][0],
                "factor": curr_image_data[3][1]
            }
            gallery_json["content"]["annotations"][curr_image_name][zoom_to_figure_name] = zoom_params

        return gallery_json

    def to_json(self):
        annotations = {}
        layout = [[] for _ in range(self.col_number)]
        index_in_layout = 0

        for curr_data_name, curr_url_ann_index in self._data.items():
            annotations[curr_data_name] = self._get_item_annotation(curr_data_name)

            curr_col_index = curr_url_ann_index[2]
            if curr_col_index is not None:
                layout[curr_col_index - 1].append(curr_data_name)
            else:
                if index_in_layout == self.col_number:
                    index_in_layout = 0
                layout[index_in_layout].append(curr_data_name)
                index_in_layout += 1

        if self._need_zoom:
            items = self._data.items()
            zoom_to_figure_name = "zoomToFigure"
            for item in items:
                curr_image_name = item[0]
                curr_image_data = item[1]

                if len(curr_image_data) < 4:
                    raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))

                elif type(curr_image_data[3]) is not tuple:
                    raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))

                zoom_params = {
                    "figureId": curr_image_data[3][0],
                    "factor": curr_image_data[3][1]
                }
                annotations[curr_image_name][zoom_to_figure_name] = zoom_params

        return {
            "content": {
                "projectMeta": self._project_meta.to_json(),
                "layout": layout,
                "annotations": annotations
            },
            "options": {
                **self._options
            }
        }


# coding: utf-8

import os.path
from pkg_resources import parse_version
import base64
import requests
import numpy as np

import cv2
from PIL import ImageDraw, ImageFile, ImageFont, Image as PILImage
import numpy as np
from enum import Enum
import skimage.transform

from supervisely_lib.io.fs import ensure_base_path, get_file_ext
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.image_rotator import ImageRotator
from supervisely_lib.imaging.font import get_font
from supervisely_lib._utils import get_bytes_hash

ImageFile.LOAD_TRUNCATED_IMAGES = True

#@TODO: refactoring image->img
KEEP_ASPECT_RATIO = -1  # TODO: need move it to best place

# Do NOT use directly for image extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
SUPPORTED_IMG_EXTS = ['.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp']
DEFAULT_IMG_EXT = '.png'


class CornerAnchorMode:
    TOP_LEFT = "tl"
    TOP_RIGHT = "tr"
    BOTTOM_LEFT = 'bl'
    BOTTOM_RIGHT = 'br'


class RotateMode(Enum):
    KEEP_BLACK = 0
    CROP_BLACK = 1
    SAVE_ORIGINAL_SIZE = 2


class ImageExtensionError(Exception):
    pass


class UnsupportedImageFormat(Exception):
    pass


class ImageReadException(Exception):
    pass

def is_valid_ext(ext: str) -> bool:
    '''
    The function is_valid_ext checks file extension for list of supported images extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
    :param ext: file extention
    :return: True if file extention in list of supported images extensions, False - in otherwise
    '''
    return ext.lower() in SUPPORTED_IMG_EXTS


def has_valid_ext(path: str) -> bool:
    '''
    The function has_valid_ext checks if a given file has a supported extension('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
    :param path: the path to the input file
    :return: True if a given file has a supported extension, False - in otherwise
    '''
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(path):
    '''
    The function validate_ext generate exception error if file extention is not in list of supported images
    extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
    :param path: the path to the input file
    '''
    _, ext = os.path.splitext(path)
    if not is_valid_ext(ext):
        raise ImageExtensionError(
            'Unsupported image extension: {!r} for file {!r}. Only the following extensions are supported: {}.'.format(
                ext, path, ', '.join(SUPPORTED_IMG_EXTS)))


def validate_format(path):
    '''
    The function validate_format checks input file format. It generate exception error(ImageReadException) if error
    has occured trying to read image and generate exception error(UnsupportedImageFormat) if file extention is not in
    list of supported images extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
    :param path: the path to the input file
    '''
    try:
        pil_img = PILImage.open(path)
        pil_img.load()  # Validate image data. Because 'open' is lazy method.
    except OSError as e:
        raise ImageReadException(
            'Error has occured trying to read image {!r}. Original exception message: {!r}'.format(path, str(e)))

    img_format = pil_img.format
    img_ext = '.' + img_format
    if not is_valid_ext('.' + img_format):
        raise UnsupportedImageFormat(
            'Unsupported image format {!r} for file {!r}. Only the following formats are supported: {}'.format(
                img_ext, path, ', '.join(SUPPORTED_IMG_EXTS)))


def read(path, remove_alpha_channel=True) -> np.ndarray:
    '''
    The function read loads an image from the specified file and returns it in RGB format. If the image cannot be read
    it generate exception error(ImageReadException) if error has occured trying to read image and generate exception
    error(UnsupportedImageFormat) if file extention is not in list of supported images
    extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp')
    :param path: the path to the input file
    :return: image in RGB format(numpy matrix)
    '''
    validate_format(path)
    if remove_alpha_channel is True:
        img = cv2.imread(path, cv2.IMREAD_COLOR)
        if img is None:
            raise IOError("OpenCV can not open the file {!r}".format(path))
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    else:
        img = cv2.imread(path, cv2.IMREAD_UNCHANGED)
        if img is None:
            raise IOError("OpenCV can not open the file {!r}".format(path))
        cnt_channels = img.shape[2]
        if cnt_channels == 4:
            return cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
        elif cnt_channels == 3:
            return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif cnt_channels == 1:
            return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            raise ValueError("image has {} channels. Please, contact support...".format(cnt_channels))


def read_bytes(image_bytes, keep_alpha=False) -> np.ndarray:
    '''
    The function read_bytes loads an byte image and returns it in RGB format.
    :param image_bytes: byte image
    :return: image in RGB format(numpy matrix)
    '''
    image_np_arr = np.asarray(bytearray(image_bytes), dtype="uint8")
    if keep_alpha is True:
        img = cv2.imdecode(image_np_arr, cv2.IMREAD_UNCHANGED)
        if len(img.shape) == 2:
            img = np.expand_dims(img, 2)
        cnt_channels = img.shape[2]
        if cnt_channels == 4:
            img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
        elif cnt_channels == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif cnt_channels == 1:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img
    else:
        img = cv2.imdecode(image_np_arr, cv2.IMREAD_COLOR)  # cv2.imdecode returns BGR always
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)


def write(path, img, remove_alpha_channel=True):
    '''
    The function write saves the image to the specified file. It create directory from path if the directory for this
    path does not exist. It generate exception error(UnsupportedImageFormat) if file extention is not in list
    of supported images extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp').
    :param path: the path to the output file
    :param img: image in RGB format(numpy matrix)
    '''
    ensure_base_path(path)
    validate_ext(path)
    res_img = img.copy()
    if len(img.shape) == 2:
        res_img = np.expand_dims(img, 2)
    cnt_channels = res_img.shape[2]
    if cnt_channels == 4:
        if remove_alpha_channel is True:
            res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGBA2BGR)
        else:
            res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGBA2BGRA)
    elif cnt_channels == 3:
        res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGB2BGR)
    elif cnt_channels == 1:
        res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_GRAY2BGR)
    return cv2.imwrite(path, res_img)


def draw_text_sequence(bitmap: np.ndarray,
                       texts: list,
                       anchor_point: tuple,
                       corner_snap: CornerAnchorMode = CornerAnchorMode.TOP_LEFT,
                       col_space: int = 12,
                       font: ImageFont.FreeTypeFont = None,
                       fill_background: bool = True) -> None:
    """
    Draws text labels on bitmap from left to right with `col_space` spacing between labels.

    Args:
        bitmap: target image (canvas)
        texts: texts sequence for drawing
        anchor_point: start anchor point (row, column)
        corner_snap: control, how to draw text around `anchor_point`
        col_space: horizontal space between text labels in pixels.
        font: True-Type font object
        fill_background: draw background or not.
    """
    col_offset = 0
    for text in texts:
        position = anchor_point[0], anchor_point[1] + col_offset
        _, text_width = draw_text(bitmap, text, position, corner_snap, font, fill_background)
        col_offset += text_width + col_space


def draw_text(bitmap: np.ndarray,
              text: str,
              anchor_point: tuple,
              corner_snap: CornerAnchorMode=CornerAnchorMode.TOP_LEFT,
              font: ImageFont.FreeTypeFont = None,
              fill_background=True) -> tuple:
    """
    Draws given text on bitmap image.
    Args:
        bitmap: target image (canvas)
        text: text for drawing
        anchor_point: start anchor point (row, column)
        corner_snap: control, how to draw text around `anchor_point`
        font: True-Type font object
        fill_background: draw background or not.

    Returns:
        Calculated (text_height, text_width) tuple. It may be helpful for some calculations
    """

    if font is None:
        font = get_font()

    source_img = PILImage.fromarray(bitmap)
    source_img = source_img.convert("RGBA")

    canvas = PILImage.new('RGBA', source_img.size, (0, 0, 0, 0))
    drawer = ImageDraw.Draw(canvas, "RGBA")
    text_width, text_height = drawer.textsize(text, font=font)
    rect_top, rect_left = anchor_point

    if corner_snap == CornerAnchorMode.TOP_LEFT:
        pass  # Do nothing
    elif corner_snap == CornerAnchorMode.TOP_RIGHT:
        rect_left -= text_width
    elif corner_snap == CornerAnchorMode.BOTTOM_LEFT:
        rect_top -= text_height
    elif corner_snap == CornerAnchorMode.BOTTOM_RIGHT:
        rect_top -= text_height
        rect_left -= text_width

    if fill_background:
        rect_right = rect_left + text_width
        rect_bottom = rect_top + text_height
        drawer.rectangle(((rect_left, rect_top), (rect_right+1, rect_bottom)), fill=(255, 255, 255, 128))
    drawer.text((rect_left+1, rect_top), text, fill=(0, 0, 0, 255), font=font)

    source_img = PILImage.alpha_composite(source_img, canvas)
    source_img = source_img.convert("RGB")
    bitmap[:,:,:] = np.array(source_img, dtype=np.uint8)

    return (text_height, text_width)


#@TODO: not working with alpha channel
def write_bytes(img, ext) -> np.ndarray:
    '''
    The function compresses the image and stores it in the byte object. It generate exception
    error(UnsupportedImageFormat) if file extention is not in list of supported images
    extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp'). Function generate exception error(RuntimeError) if it
    can't encode input image.
    :param img: image in RGB format(numpy matrix) to be written
    :param ext: file extension that defines the output format
    :return: the byte object
    '''
    ext = ('.' + ext).replace('..', '.')
    if not is_valid_ext(ext):
        raise UnsupportedImageFormat(
            'Unsupported image format {!r}. Only the following formats are supported: {}'.format(
                ext, ', '.join(SUPPORTED_IMG_EXTS)))
    img = cv2.cvtColor(img.astype(np.uint8), cv2.COLOR_RGB2BGR)
    encode_status, img_array = cv2.imencode(ext, img)
    if encode_status is True:
        return img_array.tobytes()
    raise RuntimeError('Can not encode input image')


#@TODO: not working with alpha channel
def get_hash(img, ext):
    '''
    The function get_hash hash input image with sha256 algoritm and encode result by using Base64
    :param img: image in RGB format(numpy matrix)
    :param ext: file extension that defines the output format
    :return: the byte object
    '''
    return get_bytes_hash(write_bytes(img, ext))


def crop(img: np.ndarray, rect: Rectangle) -> np.ndarray:
    '''
    The function crop cut out part of the image with rectangle size. If rectangle for crop is out of image area it
    generates exception error(ValueError).
    :param img: image(numpy matrix) to be cropped
    :param rect: class object Rectangle of a certain size
    :return: cropped image
    '''
    img_rect = Rectangle.from_array(img)
    if not img_rect.contains(rect):
        raise ValueError('Rectangle for crop out of image area!')
    return rect.get_cropped_numpy_slice(img)


def crop_with_padding(img: np.ndarray, rect: Rectangle) -> np.ndarray:
    '''
    The function crop cut out part of the image with rectangle size. If rectangle for crop is out of image area it
    generates additional padding.
    :param img: image(numpy matrix) to be cropped
    :param rect: class object Rectangle of a certain size
    :return: cropped image
    '''
    img_rect = Rectangle.from_array(img)
    if not img_rect.contains(rect):
        row, col = img.shape[:2]
        new_img = cv2.copyMakeBorder(
            img,
            top=rect.height,
            bottom=rect.height,
            left=rect.width,
            right=rect.width,
            borderType=cv2.BORDER_CONSTANT
        )
        new_rect = rect.translate(drow=rect.height, dcol=rect.width)
        return new_rect.get_cropped_numpy_slice(new_img)

    else:
        return rect.get_cropped_numpy_slice(img)


def restore_proportional_size(in_size: tuple, out_size: tuple = None,
                              frow: float = None, fcol: float = None, f: float = None):
    '''
    The function restore_proportional_size calculate new size of the image. If some parameters are not specified, or
    size dimensions are less than 0 it generate exception error(ValueError).
    :param in_size: size of input image
    :param out_size: new image size
    :param frow: length of output image
    :param fcol: height of output image
    :param f: positive non zero scale factor
    :return: size of output image
    '''
    if out_size is not None and (frow is not None or fcol is not None) and f is None:
        raise ValueError('Must be specified output size or scale factors not both of them.')

    if out_size is not None:
        if out_size[0] == KEEP_ASPECT_RATIO and out_size[1] == KEEP_ASPECT_RATIO:
            raise ValueError('Must be specified at least 1 dimension of size!')

        if (out_size[0] <= 0 and out_size[0] != KEEP_ASPECT_RATIO) or \
                (out_size[1] <= 0 and out_size[1] != KEEP_ASPECT_RATIO):
            raise ValueError('Size dimensions must be greater than 0.')

        result_row = out_size[0] if out_size[0] > 0 else max(1, round(out_size[1] / in_size[1] * in_size[0]))
        result_col = out_size[1] if out_size[1] > 0 else max(1, round(out_size[0] / in_size[0] * in_size[1]))
    else:
        if f is not None:
            if f < 0:
                raise ValueError('"f" argument must be positive!')
            frow = fcol = f

        if (frow < 0 or fcol < 0) or (frow is None or fcol is None):
            raise ValueError('Specify "f" argument for single scale!')

        result_col = round(fcol * in_size[1])
        result_row = round(frow * in_size[0])
    return result_row, result_col


#@TODO: reimplement, to be more convenient
def resize(img: np.ndarray, out_size: tuple=None, frow: float=None, fcol: float=None) -> np.ndarray:
    '''
    The function resize resizes the image img down to or up to the specified size. If some parameters are not specified, or
    size dimensions are less than 0 it generate exception error(ValueError).
    :param img: input image
    :param out_size: new image size
    :param frow: length of output image
    :param fcol: height of output image
    :return: resized image
    '''
    result_height, result_width = restore_proportional_size(img.shape[:2], out_size, frow, fcol)
    return cv2.resize(img, (result_width, result_height), interpolation=cv2.INTER_CUBIC)


def resize_inter_nearest(img: np.ndarray, out_size: tuple=None, frow: float=None, fcol: float=None) -> np.ndarray:
    '''
    The function resize_inter_nearest resize image to match a certain size. Performs interpolation to up-size or
    down-size images. For down-sampling N-dimensional images by applying a function or the arithmetic mean, see
    `skimage.measure.block_reduce` and `skimage.transform.downscale_local_mean`, respectively.
    If some parameters are not specified, or size dimensions are less than 0 it generate exception
    error(ValueError).
    :param img: input image
    :param out_size: new image size
    :param frow: length of output image
    :param fcol: height of output image
    :return: resized image
    '''
    target_shape = restore_proportional_size(img.shape[:2], out_size, frow, fcol)
    resize_kv_args = dict(order=0, preserve_range=True, mode='constant')
    if parse_version(skimage.__version__) >= parse_version('0.14.0'):
        resize_kv_args['anti_aliasing'] = False
    return skimage.transform.resize(img, target_shape, **resize_kv_args).astype(img.dtype)


def scale(img: np.ndarray, factor: float) -> np.ndarray:
    """
    :param factor: positive non zero scale factor
    :return: scaled by factor image
    """
    return resize(img, (round(img.shape[0] * factor), round(img.shape[1] * factor)))


def fliplr(img: np.ndarray) -> np.ndarray:
    """
    :return: image flipped horizontally (columns order reversed)
    """
    return np.flip(img, 1)


def flipud(img: np.ndarray) -> np.ndarray:
    """
    :return: image flipped vertically (rows order reversed)
    """
    return np.flip(img, 0)


def rotate(img: np.ndarray, degrees_angle: float, mode=RotateMode.KEEP_BLACK) -> np.ndarray:
    """
    Rotate given a NumPy / OpenCV image on selected angle and Extend/Crop by chosen mode.
    :param img: image for rotation
    :param degrees_angle: angle in degrees
    :param mode: one of RotateMode enum values
    :return: rotated and processed image
    """

    rotator = ImageRotator(imsize=img.shape[:2], angle_degrees_ccw=degrees_angle)
    if mode == RotateMode.KEEP_BLACK:
        return rotator.rotate_img(img, use_inter_nearest=False)  # @TODO: order = ???
    elif mode == RotateMode.CROP_BLACK:
        img_rotated = rotator.rotate_img(img, use_inter_nearest=False)
        return rotator.inner_crop.get_cropped_numpy_slice(img_rotated)
    elif mode == RotateMode.SAVE_ORIGINAL_SIZE:
        # TODO Implement this in rotator instead.
        return skimage.transform.rotate(img, degrees_angle, resize=False)
    else:
        raise NotImplementedError('Rotate mode "{0}" not supported!'.format(str(mode)))


# Color augmentations
def _check_contrast_brightness_inputs(min_value, max_value):
    '''
    The function _check_contrast_brightness_inputs checks the input brightness or contrast for correctness.
    '''
    if min_value < 0:
        raise ValueError('Minimum value must be greater than or equal to 0.')
    if min_value > max_value:
        raise ValueError('Maximum value must be greater than or equal to minimum value.')


def random_contrast(image: np.ndarray, min_factor: float, max_factor: float) -> np.ndarray:
    """
    Randomly changes contrast of the input image.

    Args:
        image: Input image array.
        min_factor: Lower bound of contrast range.
        max_factor: Upper bound of contrast range.
    Returns:
        Image array with changed contrast.
    """
    _check_contrast_brightness_inputs(min_factor, max_factor)
    contrast_value = np.random.uniform(min_factor, max_factor)
    image_gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    image_mean = round(image_gray.mean())
    image = image.astype(np.float32)
    image = contrast_value * (image - image_mean) + image_mean
    return np.clip(image, 0, 255).astype(np.uint8)


def random_brightness(image: np.ndarray, min_factor: float, max_factor: float) -> np.ndarray:
    """
    Randomly changes brightness of the input image.

    Args:
        image: Input image array.
        min_factor: Lower bound of brightness range.
        max_factor: Upper bound of brightness range.
    Returns:
        Image array with changed brightness.
    """
    _check_contrast_brightness_inputs(min_factor, max_factor)
    brightness_value = np.random.uniform(min_factor, max_factor)
    image = image.astype(np.float32)
    image = image * brightness_value
    return np.clip(image, 0, 255).astype(np.uint8)


def random_noise(image: np.ndarray, mean: float, std: float) -> np.ndarray:
    """
    Adds random Gaussian noise to the input image.

    Args:
        image: Input image array.
        mean: The mean value of noise distribution.
        std: The standard deviation of noise distribution.
    Returns:
        Image array with additional noise.
    """
    image = image.astype(np.float32)
    image += np.random.normal(mean, std, image.shape)

    return np.clip(image, 0, 255).astype(np.uint8)


def random_color_scale(image: np.ndarray, min_factor: float, max_factor: float) -> np.ndarray:
    """
    Changes image colors by randomly scaling each of RGB components. The scaling factors are sampled uniformly from
    the given range.
    Args:
        image: Input image array.
        min_factor: minimum scale factor
        max_factor: maximum scale factor
    Returns:
        Image array with shifted colors.
    """
    image_float = image.astype(np.float64)
    scales = np.random.uniform(low=min_factor, high=max_factor, size=(1,1, image.shape[2]))
    res_image = image_float * scales
    return np.clip(res_image, 0, 255).astype(np.uint8)


# Blurs
def blur(image: np.ndarray, kernel_size: int) -> np.ndarray:
    """
    Blurs an image using the normalized box filter.

    Args:
        image: Input image array.
        kernel_size: Blurring kernel size.
    Returns:
        Blurred image array.
    """
    return cv2.blur(image, (kernel_size, kernel_size))


def median_blur(image: np.ndarray, kernel_size: int) -> np.ndarray:
    """
    Blurs an image using the median filter.

    Args:
        image: Input image array.
        kernel_size: Blurring kernel size.
    Returns:
        Blurred image array.
    """
    return cv2.medianBlur(image, kernel_size)


def gaussian_blur(image: np.ndarray, sigma_min: float, sigma_max: float) -> np.ndarray:
    """
    Blurs an image using a Gaussian filter.

    Args:
        image: Input image array.
        sigma_min: Lower bound of Gaussian kernel standard deviation range.
        sigma_max: Upper bound of Gaussian kernel standard deviation range.
    Returns:
        Blurred image array.
    """
    sigma_value = np.random.uniform(sigma_min, sigma_max)
    return cv2.GaussianBlur(image, (0, 0), sigma_value)


def drop_image_alpha_channel(img: np.ndarray) -> np.ndarray:
    """
    Converts 4-channel image to 3-channel.

    Args:
        img: Input 4-channel image array.
    Returns:
        3-channel image array.
    """
    if img.shape[2] != 4:
        raise ValueError('Only 4-channel RGBA images are supported for alpha channel removal. ' +
                         'Instead got {} channels.'.format(img.shape[2]))
    return cv2.cvtColor(img, cv2.COLOR_RGBA2RGB)


#@TODO: refactor from two separate methods to a single one
#bgra or bgr
def np_image_to_data_url(img):
    encode_status, bgra_result_png = cv2.imencode('.png', img)
    img_png = bgra_result_png.tobytes()
    img_base64 = base64.b64encode(img_png)
    data_url = 'data:image/png;base64,{}'.format(str(img_base64, 'utf-8'))
    return data_url


def data_url_to_numpy(data_url):
    img_base64 = data_url[len('data:image/png;base64,'):]
    img_base64 = base64.b64decode(img_base64)
    image = read_bytes(img_base64)
    return image


#only rgb
def np_image_to_data_url_backup_rgb(img):
    img_base64 = base64.b64encode(write_bytes(img, 'png'))
    data_url = 'data:image/png;base64,{}'.format(str(img_base64, 'utf-8'))
    return data_url


# coding: utf-8

from supervisely_lib.api.module_api import ModuleApi
from supervisely_lib.api.module_api import ApiField


class ProjectClassApi(ModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                #ApiField.PROJECT_ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.SHAPE,
                ApiField.COLOR,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT
                ]

    @staticmethod
    def info_tuple_name():
        return 'ProjectClassInfo'

    def get_list(self, project_id, filters=None):
        '''
        :param project_id: int
        :param filters: list
        :return: list all the classes for a given project
        '''
        return self.get_list_all_pages('advanced.object_classes.list',  {ApiField.PROJECT_ID: project_id, "filter": filters or []})



# coding: utf-8

from supervisely_lib.api.module_api import ApiField, ModuleApiBase


class AdvancedApi(ModuleApiBase):
    def add_tag_to_object(self, tag_meta_id, figure_id, value=None):
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id}
        if value is not None:
            data[ApiField.VALUE] = value
        resp = self._api.post('object-tags.add-to-object', data)
        return resp.json()

    def remove_tag_from_object(self, tag_meta_id, figure_id, tag_id):
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.FIGURE_ID: figure_id, ApiField.ID: tag_id}
        resp = self._api.post('object-tags.remove-from-figure', data)
        return resp.json()

    def get_object_tags(self, figure_id):
        data = {ApiField.ID: figure_id}
        resp = self._api.post('figures.tags.list', data)
        return resp.json()

    def remove_tag_from_image(self, tag_meta_id, image_id, tag_id):
        data = {ApiField.TAG_ID: tag_meta_id, ApiField.IMAGE_ID: image_id, ApiField.ID: tag_id}
        resp = self._api.post('image-tags.remove-from-image', data)
        return resp.json()



# coding: utf-8

from supervisely_lib.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID, INDICES
from supervisely_lib.geometry.geometry import Geometry


class Pointcloud(Geometry):
    @staticmethod
    def geometry_name():
        return 'point_cloud'

    def __init__(self, indices, sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        if type(indices) is not list:
            raise TypeError("\"indices\" param has to be of type {!r}".format(type(list)))

        self._indices = indices

    @property
    def indices(self):
        return self._indices.copy()

    def to_json(self):
        res = {INDICES: self.indices}
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        indices = data[INDICES]

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(indices, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)



# coding: utf-8
from collections import namedtuple

from supervisely_lib.annotation.label import Label
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.metric.common import safe_ratio

import numpy as np


LabelsPairWithScore = namedtuple('LabelsPairWithScore', ['label_1', 'label_2', 'score'])
LabelsMatchResult = namedtuple('LabelsMatchResult', ['matches', 'unmatched_labels_1', 'unmatched_labels_2'])
IndexPairWithScore = namedtuple('IndexPairWithScore', ['idx_1', 'idx_2', 'score'])
IndexMatchResult = namedtuple('IndexMatchResult', ['matches', 'unmatched_indices_1', 'unmatched_indices_2'])


def filter_labels_by_name(labels, names_whitelist):
    return [label for label in labels if label.obj_class.name in names_whitelist]


def get_iou_rect(rect: Rectangle, other: Geometry):
    maybe_other_cropped = other.crop(rect)
    if len(maybe_other_cropped) == 0:
        return 0.0
    else:
        [other_cropped] = maybe_other_cropped
        intersection_area = other_cropped.area
        if intersection_area == 0:
            return 0.0
        union_area = rect.area + other.area - intersection_area
        return intersection_area / union_area


def get_geometries_iou(geometry_1: Geometry, geometry_2: Geometry):
    if isinstance(geometry_1, Rectangle):
        return get_iou_rect(geometry_1, geometry_2)
    elif isinstance(geometry_2, Rectangle):
        return get_iou_rect(geometry_2, geometry_1)
    else:
        common_bbox = Rectangle.from_geometries_list((geometry_1, geometry_2))
        g1 = geometry_1.relative_crop(common_bbox)[0]
        g2 = geometry_2.relative_crop(common_bbox)[0]
        mask_1 = np.full(common_bbox.to_size(), False)
        g1.draw(mask_1, color=True)
        mask_2 = np.full(common_bbox.to_size(), False)
        g2.draw(mask_2, color=True)
        return safe_ratio((mask_1 & mask_2).sum(), (mask_1 | mask_2).sum())


def get_labels_iou(label_1: Label, label_2: Label, img_size=None):
    return get_geometries_iou(label_1.geometry, label_2.geometry)


def match_indices_by_score(elems_1, elems_2, score_threshold, score_fn):
    # Score all the possible pairs.
    scored_idx_pairs = [
        IndexPairWithScore(idx_1=idx_1, idx_2=idx_2, score=score_fn(elem_1, elem_2))
        for idx_1, elem_1 in enumerate(elems_1) for idx_2, elem_2 in enumerate(elems_2)]
    # Apply the threshold to avoid sorting candidates with too low scores.
    thresholded_idx_pairs = [p for p in scored_idx_pairs if p.score >= score_threshold]
    # Sort by score in descending order.
    sorted_idx_pairs = sorted(thresholded_idx_pairs, key=lambda p: p.score, reverse=True)

    # Match greedily, make sure no element is matched to more than one counterpart.
    unmatched_1 = set(range(len(elems_1)))
    unmatched_2 = set(range(len(elems_2)))
    matches = []
    for idx_pair in sorted_idx_pairs:
        if idx_pair.idx_1 in unmatched_1 and idx_pair.idx_2 in unmatched_2:
            matches.append(idx_pair)
            unmatched_1.remove(idx_pair.idx_1)
            unmatched_2.remove(idx_pair.idx_2)
    return IndexMatchResult(matches=matches, unmatched_indices_1=unmatched_1, unmatched_indices_2=unmatched_2)


def match_labels_by_iou(labels_1, labels_2, img_size, iou_threshold):
    index_matches = match_indices_by_score(labels_1, labels_2, iou_threshold, score_fn=get_labels_iou)
    return LabelsMatchResult(
        matches=[LabelsPairWithScore(label_1=labels_1[match.idx_1], label_2=labels_2[match.idx_2], score=match.score)
                 for match in index_matches.matches],
        unmatched_labels_1=[labels_1[idx] for idx in index_matches.unmatched_indices_1],
        unmatched_labels_2=[labels_2[idx] for idx in index_matches.unmatched_indices_2])



# coding: utf-8

import uuid
from bidict import bidict

from supervisely_lib.annotation.label import LabelJsonFields
from supervisely_lib.annotation.obj_class import ObjClass
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib._utils import take_with_default
from supervisely_lib.video_annotation.constants import KEY, ID, OBJECTS_MAP
from supervisely_lib.video_annotation.video_tag_collection import VideoTagCollection
from supervisely_lib.video_annotation.video_tag import VideoTag
from supervisely_lib.collection.key_indexed_collection import KeyObject
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, CLASS_ID


class VideoObject(KeyObject):
    '''
    This is a class for creating and using VideoObject objects for videos
    '''
    def __init__(self, obj_class: ObjClass, tags: VideoTagCollection = None, key=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param obj_class: ObjClass class object
        :param tags: VideoTagCollection
        :param key: uuid class object
        '''
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at

        self._obj_class = obj_class
        self._key = take_with_default(key, uuid.uuid4())
        self._tags = take_with_default(tags, VideoTagCollection())

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at

    @property
    def obj_class(self):
        return self._obj_class

    def key(self):
        return self._key

    @property
    def tags(self):
        return self._tags.clone()

    def add_tag(self, tag: VideoTag):
        '''
        Add given tag to VideoTagCollection
        :param tag: VideoTag class object
        :return: VideoObject class object
        '''
        return self.clone(tags=self._tags.add(tag))

    def add_tags(self, tags: list):
        '''
        Add given tags to VideoTagCollection
        :param tags: list of VideoTag class objects
        :return: VideoObject class object
        '''
        return self.clone(tags=self._tags.add_items(tags))

    def to_json(self, key_id_map: KeyIdMap = None):
        '''
        The function to_json convert VideoObject class object to json format
        :param key_id_map: KeyIdMap class object
        :return: VideoObject in json format
        '''
        data_json = {
            KEY: self.key().hex,
            LabelJsonFields.OBJ_CLASS_NAME: self.obj_class.name,
            LabelJsonFields.TAGS: self.tags.to_json(key_id_map)
        }

        if key_id_map is not None:
            item_id = key_id_map.get_object_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        self._add_creation_info(data_json)
        return data_json

    @classmethod
    def from_json(cls, data, project_meta: ProjectMeta, key_id_map: KeyIdMap = None):
        '''
        The function from_json convert VideoObject from json format to VideoObject class object. Raise error if object class name is not found in the given project meta
        :param data: input VideoObject in json format
        :param project_meta: ProjectMeta class object
        :param key_id_map: KeyIdMap class object
        :return: VideoObject class object
        '''
        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(f'Failed to deserialize a object from JSON: class name {obj_class_name!r} '
                               f'was not found in the given project meta.')

        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_object(key, data.get(ID, None))

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(obj_class=obj_class,
                   key=key,
                   tags=VideoTagCollection.from_json(data[LabelJsonFields.TAGS], project_meta.tag_metas),
                   class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def clone(self, obj_class: ObjClass=None, tags: VideoTagCollection = None, key=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param obj_class: ObjClass
        :param tags: VideoTagCollection
        :param key: uuid class object
        :return: VideoObject class object
        '''
        return self.__class__(obj_class=take_with_default(obj_class, self.obj_class),
                              key=take_with_default(key, self._key),
                              tags=take_with_default(tags, self.tags),
                              class_id=take_with_default(class_id, self.class_id),
                              labeler_login=take_with_default(labeler_login, self.labeler_login),
                              updated_at=take_with_default(updated_at, self.updated_at),
                              created_at=take_with_default(created_at, self.created_at))




# coding: utf-8

from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection
from supervisely_lib.api.entity_annotation.object_api import ObjectApi


class PointcloudObjectApi(ObjectApi):
    def append_bulk(self, pointcloud_id, objects: PointcloudObjectCollection, key_id_map: KeyIdMap = None):
        info = self._api.pointcloud.get_info_by_id(pointcloud_id)
        return self._append_bulk(self._api.pointcloud.tag, pointcloud_id, info.project_id, info.dataset_id, objects, key_id_map, is_pointcloud=True)

    def append_to_dataset(self, dataset_id, objects: PointcloudObjectCollection, key_id_map: KeyIdMap = None):
        project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
        return self._append_bulk(self._api.pointcloud.tag, dataset_id, project_id, dataset_id, objects, key_id_map, is_pointcloud=True)



# coding: utf-8

from supervisely_lib.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib._utils import batched


class FigureApi(RemoveableBulkModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.UPDATED_AT,
                ApiField.CREATED_AT,
                ApiField.ENTITY_ID,
                ApiField.OBJECT_ID,
                ApiField.PROJECT_ID,
                ApiField.DATASET_ID,
                ApiField.FRAME_INDEX,
                ApiField.GEOMETRY_TYPE,
                ApiField.GEOMETRY
                ]

    @staticmethod
    def info_tuple_name():
        return 'FigureInfo'

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: figure metadata by numeric id
        '''
        return self._get_info_by_id(id, 'figures.info')

    def create(self, entity_id, object_id, meta, geometry_json, geometry_type, track_id=None):
        input_figure = {
                    ApiField.META: meta,
                    ApiField.OBJECT_ID: object_id,
                    ApiField.GEOMETRY_TYPE: geometry_type,
                    ApiField.GEOMETRY: geometry_json};

        if track_id is not None:
            input_figure[ApiField.TRACK_ID] = track_id

        body = {ApiField.ENTITY_ID: entity_id,
                ApiField.FIGURES: [input_figure]}

        response = self._api.post('figures.bulk.add', body)
        return response.json()[0][ApiField.ID]

    def get_by_ids(self, dataset_id, ids):
        '''
        :param dataset_id: int
        :param ids: list of integers
        :return: list of figures with given ids from dataset with given id
        '''
        filters = [{"field": "id", "operator": "in", "value": ids}]
        figures_infos = self.get_list_all_pages('figures.list', {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters})

        if len(ids) != len(figures_infos):
            ids_downloaded = [info.id for info in figures_infos]
            raise RuntimeError("Ids don't exist on server: {}".format(set(ids_downloaded) - set(ids)))

        id_to_item = {info.id: info for info in figures_infos}

        figures = []
        for input_id in ids:
            figures.append(id_to_item[input_id])

        return figures

    def _append_bulk(self, entity_id, figures_json, figures_keys, key_id_map: KeyIdMap, field_name=ApiField.ENTITY_ID):
        if len(figures_json) == 0:
            return

        for (batch_keys, batch_jsons) in zip(batched(figures_keys, batch_size=100),
                                             batched(figures_json, batch_size=100)):
            resp = self._api.post('figures.bulk.add', {field_name: entity_id, ApiField.FIGURES: batch_jsons})
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)



# coding: utf-8

from copy import deepcopy

import numpy as np
from supervisely_lib.io.json import JsonSerializable
from supervisely_lib.geometry.constants import ANY_SHAPE, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely_lib import logger


# @TODO: use properties instead of field if it makes sense
class Geometry(JsonSerializable):
    def __init__(self, sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        self.sly_id = sly_id
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at
        self.class_id = class_id

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at
        if self.sly_id is not None:
            d[ID] = self.sly_id
        if self.class_id is not None:
            d[CLASS_ID] = self.class_id

    def _copy_creation_info_inplace(self, g):
        self.labeler_login = g.labeler_login
        self.updated_at = g.updated_at
        self.created_at = g.created_at
        self.sly_id = g.sly_id

    @staticmethod
    def geometry_name():
        """
        :return: string with name of geometry
        """
        raise NotImplementedError()

    def crop(self, rect):
        """
        :param rect: Rectangle
        :return: list of Geometry
        """
        raise NotImplementedError()

    def relative_crop(self, rect):
        """ Crops object like "crop" method, but return results with coordinates relative to rect
        :param rect:
        :return: list of Geometry
        """
        return [geom.translate(drow=-rect.top, dcol=-rect.left) for geom in self.crop(rect)]

    def rotate(self, rotator):
        """Rotates around image center -> New Geometry
        :param rotator: ImageRotator
        :return: Geometry
        """
        raise NotImplementedError()

    def resize(self, in_size, out_size):
        """
        :param in_size: (rows, cols)
        :param out_size:
            (128, 256)
            (128, KEEP_ASPECT_RATIO)
            (KEEP_ASPECT_RATIO, 256)
        :return: Geometry
        """
        raise NotImplementedError()

    def scale(self, factor):
        """Scales around origin with a given factor.
        :param: factor (float):
        :return: Geometry
        """
        raise NotImplementedError()

    def translate(self, drow, dcol):
        """
        :param drow: int rows shift
        :param dcol: int cols shift
        :return: Geometry
        """
        raise NotImplementedError()

    def fliplr(self, img_size):
        """
        :param img_size: (rows, cols)
        :return: Geometry
        """
        raise NotImplementedError()

    def flipud(self, img_size):
        """
        :param img_size: (rows, cols)
        :return: Geometry
        """
        raise NotImplementedError()

    def _draw_bool_compatible(self, draw_fn, bitmap, color, thickness, config=None):
        if bitmap.dtype == np.bool:
            # Cannot draw on the canvas directly, create a temporary with different type.
            temp_bitmap = np.zeros(bitmap.shape[:2], dtype=np.uint8)
            draw_fn(temp_bitmap, 1, thickness=thickness, config=config)
            bitmap[temp_bitmap == 1] = color
        else:
            # Pass through the canvas without temp bitmap for efficiency.
            draw_fn(bitmap, color, thickness=thickness, config=config)

    def draw(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: used only in Polyline and Point
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        """
        self._draw_bool_compatible(self._draw_impl, bitmap, color, thickness, config)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: used only in Polyline and Point
        """
        raise NotImplementedError()

    def draw_contour(self, bitmap, color, thickness=1, config=None):
        """ Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        """
        self._draw_bool_compatible(self._draw_contour_impl, bitmap, color, thickness, config)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        """
        raise NotImplementedError()

    @property
    def area(self):
        """
        :return: float
        """
        raise NotImplementedError()

    def to_bbox(self):
        """
        :return: Rectangle
        """
        raise NotImplementedError()

    def clone(self):
        return deepcopy(self)

    def validate(self, obj_class_shape, settings):
        if obj_class_shape != ANY_SHAPE:
            if self.geometry_name() != obj_class_shape:
                raise ValueError('Geometry validation error: shape names are mismatched!')

    @staticmethod
    def config_from_json(config):
        return config

    @staticmethod
    def config_to_json(config):
        return config

    @classmethod
    def allowed_transforms(cls):
        #raise NotImplementedError("{!r}".format(cls.geometry_name()))
        return []

    def convert(self, new_geometry, contour_radius=0, approx_epsilon=None):
        from supervisely_lib.geometry.any_geometry import AnyGeometry
        if type(self) == new_geometry or new_geometry == AnyGeometry:
            return [self]

        allowed_transforms = self.allowed_transforms()
        if new_geometry not in allowed_transforms:
            raise NotImplementedError("from {!r} to {!r}".format(self.geometry_name(), new_geometry.geometry_name()))

        from supervisely_lib.geometry.bitmap import Bitmap
        from supervisely_lib.geometry.rectangle import Rectangle
        from supervisely_lib.geometry.polygon import Polygon
        from supervisely_lib.geometry.helpers import geometry_to_bitmap, geometry_to_polygon

        res = []
        if new_geometry == Bitmap:
            res = geometry_to_bitmap(self, radius=contour_radius)
        elif new_geometry == Rectangle:
            res = [self.to_bbox()]
        elif new_geometry == Polygon:
            res = geometry_to_polygon(self, approx_epsilon=approx_epsilon)

        if len(res) == 0:
            logger.warn('Can not convert geometry {} to {} because geometry to convert is very small'.format(self.geometry_name(), new_geometry.geometry_name()))
        return res



# coding: utf-8

from supervisely_lib.collection.key_indexed_collection import MultiKeyIndexedCollection
from supervisely_lib.annotation.tag import Tag


class TagCollection(MultiKeyIndexedCollection):
    """
    Collection that stores Tag instances with unique names.
    """
    item_type = Tag

    def to_json(self):
        """
        Converts collection to json serializable list.
        Returns: json serializable dictionary
        """
        return [tag.to_json() for tag in self]

    @classmethod
    def from_json(cls, data, tag_meta_collection):
        """
        Creates collection from json serializable list.
        Returns: TagCollection
        """
        tags = [cls.item_type.from_json(tag_json, tag_meta_collection) for tag_json in data]
        return cls(tags)

    def __str__(self):
        return 'Tags:\n' + super(TagCollection, self).__str__()

    @classmethod
    def from_api_response(cls, data, tag_meta_collection, id_to_tagmeta=None):
        if id_to_tagmeta is None:
            id_to_tagmeta = tag_meta_collection.get_id_mapping()
        tags = []
        for tag_json in data:
            tag_meta_id = tag_json["tagId"]
            tag_meta = id_to_tagmeta[tag_meta_id]
            tag_json['name'] = tag_meta.name
            tag = cls.item_type.from_json(tag_json, tag_meta_collection)
            tags.append(tag)
        return cls(tags)



# coding: utf-8

import json

from supervisely_lib.annotation.annotation import Annotation
from supervisely_lib.api.module_api import ApiField, ModuleApi
from supervisely_lib._utils import batched


class AnnotationApi(ModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.IMAGE_ID,
                ApiField.IMAGE_NAME,
                ApiField.ANNOTATION,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        return 'AnnotationInfo'

    def get_list(self, dataset_id, filters=None, progress_cb=None):
        '''
        :param dataset_id: int
        :param filters: list
        :param progress_cb:
        :return: list all the annotations for a given dataset
        '''
        return self.get_list_all_pages('annotations.list',  {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters or []}, progress_cb)

    def download(self, image_id, with_custom_data=False):
        '''
        :param image_id: int
        :return: serialized JSON annotation for the image id
        '''
        response = self._api.post('annotations.info',
                                  {ApiField.IMAGE_ID: image_id, ApiField.WITH_CUSTOM_DATA: with_custom_data})
        return self._convert_json_info(response.json())

    def download_batch(self, dataset_id, image_ids, progress_cb=None, with_custom_data=False):
        '''
        :param dataset_id: int
        :param image_ids: list of integers
        :param progress_cb:
        :return: list of serialized JSON annotations for the given dataset id and image id's
        '''
        id_to_ann = {}
        for batch in batched(image_ids):
            post_data = {
                ApiField.DATASET_ID: dataset_id,
                ApiField.IMAGE_IDS: batch,
                ApiField.WITH_CUSTOM_DATA: with_custom_data
            }
            results = self._api.post('annotations.bulk.info', data=post_data).json()
            for ann_dict in results:
                ann_info = self._convert_json_info(ann_dict)
                id_to_ann[ann_info.image_id] = ann_info
            if progress_cb is not None:
                progress_cb(len(batch))
        ordered_results = [id_to_ann[image_id] for image_id in image_ids]
        return ordered_results

    def upload_path(self, img_id, ann_path):
        self.upload_paths([img_id], [ann_path])

    def upload_paths(self, img_ids, ann_paths, progress_cb=None):
        # img_ids from the same dataset
        def read_json(ann_path):
            with open(ann_path) as json_file:
                return json.load(json_file)
        self._upload_batch(read_json, img_ids, ann_paths, progress_cb)

    def upload_json(self, img_id, ann_json):
        self.upload_jsons([img_id], [ann_json])

    def upload_jsons(self, img_ids, ann_jsons, progress_cb=None):
        # img_ids from the same dataset
        self._upload_batch(lambda x: x, img_ids, ann_jsons, progress_cb)

    def upload_ann(self, img_id, ann):
        self.upload_anns([img_id], [ann])

    def upload_anns(self, img_ids, anns, progress_cb=None):
        # img_ids from the same dataset
        self._upload_batch(Annotation.to_json, img_ids, anns, progress_cb)

    def _upload_batch(self, func_ann_to_json, img_ids, anns, progress_cb=None):
        # img_ids from the same dataset
        if len(img_ids) == 0:
            return
        if len(img_ids) != len(anns):
            raise RuntimeError('Can not match "img_ids" and "anns" lists, len(img_ids) != len(anns)')

        dataset_id = self._api.image.get_info_by_id(img_ids[0]).dataset_id
        for batch in batched(list(zip(img_ids, anns))):
            data = [{ApiField.IMAGE_ID: img_id, ApiField.ANNOTATION: func_ann_to_json(ann)} for img_id, ann in batch]
            self._api.post('annotations.bulk.add', data={ApiField.DATASET_ID: dataset_id, ApiField.ANNOTATIONS: data})
            if progress_cb is not None:
                progress_cb(len(batch))

    def get_info_by_id(self, id):
        raise NotImplementedError('Method is not supported')

    def get_info_by_name(self, parent_id, name):
        raise NotImplementedError('Method is not supported')

    def exists(self, parent_id, name):
        raise NotImplementedError('Method is not supported')

    def get_free_name(self, parent_id, name):
        raise NotImplementedError('Method is not supported')

    def _add_sort_param(self, data):
        return data

    def copy_batch(self, src_image_ids, dst_image_ids, progress_cb=None):
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError('Can not match "src_image_ids" and "dst_image_ids" lists, '
                               'len(src_image_ids) != len(dst_image_ids)')
        if len(src_image_ids) == 0:
            return

        src_dataset_id = self._api.image.get_info_by_id(src_image_ids[0]).dataset_id
        for cur_batch in batched(list(zip(src_image_ids, dst_image_ids))):
            src_ids_batch, dst_ids_batch = zip(*cur_batch)
            ann_infos = self.download_batch(src_dataset_id, src_ids_batch)
            ann_jsons = [ann_info.annotation for ann_info in ann_infos]
            self.upload_jsons(dst_ids_batch, ann_jsons)
            if progress_cb is not None:
                progress_cb(len(src_ids_batch))

    def copy(self, src_image_id, dst_image_id):
        self.copy_batch([src_image_id], [dst_image_id])

    def copy_batch_by_ids(self, src_image_ids, dst_image_ids):
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError('Can not match "src_image_ids" and "dst_image_ids" lists, '
                               'len(src_image_ids) != len(dst_image_ids)')
        if len(src_image_ids) == 0:
            return

        self._api.post('annotations.bulk.copy', data={"srcImageIds": src_image_ids,
                                                      "destImageIds": dst_image_ids,
                                                      "preserveSourceDate": True})

    # def _convert_json_info(self, info: dict, skip_missing=True):
    #     res = super()._convert_json_info(info, skip_missing=skip_missing)
    #     res.annotation["imageId"] = res.image_id
    #     return res

    def append_labels(self, image_id, labels):
        if len(labels) == 0:
            return

        payload = []
        for label in labels:
            _label_json = label.to_json()
            _label_json["geometry"] = label.geometry.to_json()
            if "classId" not in _label_json:
                raise KeyError("Update project meta from server to get class id")
            payload.append(_label_json)

        added_ids = []
        for batch_jsons in batched(payload, batch_size=100):
            resp = self._api.post('figures.bulk.add', {ApiField.ENTITY_ID: image_id, ApiField.FIGURES: batch_jsons})
            for resp_obj in resp.json():
                figure_id = resp_obj[ApiField.ID]
                added_ids.append(figure_id)


import os

from setuptools import find_packages, setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fin:
        return fin.read()


# Dependencies do not include PyTorch, so
# supervisely_lib.nn.hosted.pytorch will not work out of the box.
# If you need to invoke that part of the code, it is very likely you
# already have PyTorch installed.
setup(
    name="supervisely",
    version="6.2.2",
    packages=find_packages(include=['supervisely_lib', 'supervisely_lib.*']),
    description="Supervisely Python SDK.",
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    url="https://github.com/supervisely/supervisely",
    install_requires=[
        "flask-restful>=0.3.7",
        "grpcio>=1.32.0",
        "jsonschema>=2.6.0,<3.0.0",
        "matplotlib>=3.3.2",
        "numpy>=1.19",
        "opencv-python>=3.4.10.35",
        "pandas>=1.1.3",
        "pascal-voc-writer>=0.1.4",
        "PTable>=0.9.2",
        "pillow>=5.4.1",
        "protobuf>=3.14.0",
        # Higher python-json-logger versions are incompatible with
        # simplejson somehow, so for now prevent higher versions from
        # being installed.
        "python-json-logger==0.1.11",
        "requests>=2.24.0",
        "requests-toolbelt>=0.9.1",
        "scikit-image>=0.17.1",
        "scipy>=1.5.2",
        "Shapely>=1.7.1",
        #"simplejson>=3.17.2",
        "Werkzeug>=1.0.1",
        "bidict>=0.21.2",
        "sk-video>=1.1.10",
        "plotly>=4.11.0",
        "docker>=4.4.1",
        "psutil>=5.4.5",
        "imgaug==0.4.0",
        "imagecorruptions==1.1.2"
    ],
    include_package_data=True
)



# coding: utf-8

import io
import re
import urllib.parse
import json

from requests_toolbelt import MultipartDecoder, MultipartEncoder

from supervisely_lib.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely_lib.imaging import image as sly_image
from supervisely_lib.io.fs import ensure_base_path, get_file_hash, get_file_ext, get_file_name
from supervisely_lib.sly_logger import logger
from supervisely_lib._utils import batched, generate_free_name


class ImageApi(RemoveableBulkModuleApi):
    @staticmethod
    def info_sequence():
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.LINK,
            ApiField.HASH,
            ApiField.MIME,
            ApiField.EXT,
            ApiField.SIZE,
            ApiField.WIDTH,
            ApiField.HEIGHT,
            ApiField.LABELS_COUNT,
            ApiField.DATASET_ID,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
            ApiField.META,
            ApiField.PATH_ORIGINAL,
            ApiField.FULL_STORAGE_URL,
            ApiField.TAGS,
        ]

    @staticmethod
    def info_tuple_name():
        return 'ImageInfo'

    def get_list(self, dataset_id, filters=None, sort="id", sort_order="asc"):
        '''
        :param dataset_id: int
        :param filters: list
        :param sort: string (one of "id" "name" "description" "labelsCount" "createdAt" "updatedAt")
        :return: list all the images for a given dataset
        '''
        return self.get_list_all_pages('images.list',  {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FILTER: filters or [],
            ApiField.SORT: sort,
            ApiField.SORT_ORDER: sort_order
        })

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: image metadata by numeric id
        '''
        return self._get_info_by_id(id, 'images.info')

    # @TODO: reimplement to new method images.bulk.info
    def get_info_by_id_batch(self, ids):
        '''
        :param ids: list of integers
        :return: list of images metadata
        '''
        results = []
        if len(ids) == 0:
            return results
        dataset_id = self.get_info_by_id(ids[0]).dataset_id
        for batch in batched(ids):
            filters = [{"field": ApiField.ID, "operator": "in", "value": batch}]
            results.extend(self.get_list_all_pages('images.list', {ApiField.DATASET_ID: dataset_id,
                                                                   ApiField.FILTER: filters}))
        temp_map = {info.id: info for info in results}
        ordered_results = [temp_map[id] for id in ids]
        return ordered_results

    def _download(self, id, is_stream=False):
        '''
        :param id: int
        :param is_stream: bool
        :return: Response class object contain metadata of image with given id
        '''
        response = self._api.post('images.download', {ApiField.ID: id}, stream=is_stream)
        return response

    def download_np(self, id, keep_alpha=False):
        '''
        Download image with given id in numpy format
        :param id: int
        :return: image in RGB format(numpy matrix)
        '''
        response = self._download(id)
        img = sly_image.read_bytes(response.content, keep_alpha)
        return img

    def download_path(self, id, path):
        '''
        Download image with given id and saves it for a given path
        :param id: int
        :param path: str
        '''
        response = self._download(id, is_stream=True)
        ensure_base_path(path)
        with open(path, 'wb') as fd:
            for chunk in response.iter_content(chunk_size=1024*1024):
                fd.write(chunk)

    def _download_batch(self, dataset_id, ids):
        '''
        Generate image id and it content from given dataset and list of images ids
        :param dataset_id: int
        :param ids: list of integers
        '''
        for batch_ids in batched(ids):
            response = self._api.post(
                'images.bulk.download', {ApiField.DATASET_ID: dataset_id, ApiField.IMAGE_IDS: batch_ids})
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b'Content-Disposition'].decode('utf-8')
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                img_id = int(re.findall(r'(^|[\s;])name="(\d*)"', content_utf8)[0][1])
                yield img_id, part

    def download_paths(self, dataset_id, ids, paths, progress_cb=None):
        '''
        Download images with given ids and saves it for a given paths
        :param dataset_id: int
        :param ids: list of integers
        :param paths: list of str (if ids list != lengh of paths list raise error)
        :param progress_cb:
        '''
        if len(ids) == 0:
            return
        if len(ids) != len(paths):
            raise RuntimeError("Can not match \"ids\" and \"paths\" lists, len(ids) != len(paths)")

        id_to_path = {id: path for id, path in zip(ids, paths)}
        #debug_ids = []
        for img_id, resp_part in self._download_batch(dataset_id, ids):
            #debug_ids.append(img_id)
            with open(id_to_path[img_id], 'wb') as w:
                w.write(resp_part.content)
            if progress_cb is not None:
                progress_cb(1)
        #if ids != debug_ids:
        #    raise RuntimeError("images.bulk.download: imageIds order is broken")

    def download_bytes(self, dataset_id, ids, progress_cb=None):
        '''
        Doenload images with given ids from dataset with given id in binary format(bytes)
        :param dataset_id: int
        :param ids: list of integers
        :param progress_cb:
        :return: list of images in binary format
        '''
        if len(ids) == 0:
            return []

        id_to_img = {}
        for img_id, resp_part in self._download_batch(dataset_id, ids):
            id_to_img[img_id] = resp_part.content
            if progress_cb is not None:
                progress_cb(1)

        return [id_to_img[id] for id in ids]

    def download_nps(self, dataset_id, ids, progress_cb=None, keep_alpha=False):
        '''
        Doenload images with given ids from dataset with given id in numpy format
        :param dataset_id: int
        :param ids: list of integers
        :param progress_cb:
        :return: list of images in RGB format(numpy matrix)
        '''
        return [sly_image.read_bytes(img_bytes, keep_alpha)
                for img_bytes in self.download_bytes(dataset_id=dataset_id, ids=ids, progress_cb=progress_cb)]

    def check_existing_hashes(self, hashes):
        '''
        :param hashes: list of str
        :return: list of jsons objects(None if image with given hash not exist)
        '''
        results = []
        if len(hashes) == 0:
            return results
        for hashes_batch in batched(hashes, batch_size=900):
            response = self._api.post('images.internal.hashes.list', hashes_batch)
            results.extend(response.json())
        return results

    def check_image_uploaded(self, hash):
        '''
        :param hash: str
        :return: True if image with given hash exist, False in over way
        '''
        response = self._api.post('images.internal.hashes.list', [hash])
        results = response.json()
        if len(results) == 0:
            return False
        else:
            return True

    def _upload_uniq_images_single_req(self, func_item_to_byte_stream, hashes_items_to_upload):
        """
        Upload images (binary data) to server with single request.
        Expects unique images that aren't exist at server.
        :param func_item_to_byte_stream: converter for "item" to byte stream
        :param hashes_items_to_upload: list of pairs (hash, item)
        :return: list of hashes for successfully uploaded items
        """
        content_dict = {}
        for idx, (_, item) in enumerate(hashes_items_to_upload):
            content_dict["{}-file".format(idx)] = (str(idx), func_item_to_byte_stream(item), 'image/*')
        encoder = MultipartEncoder(fields=content_dict)
        resp = self._api.post('images.bulk.upload', encoder)

        resp_list = json.loads(resp.text)
        remote_hashes = [d['hash'] for d in resp_list if 'hash' in d]
        if len(remote_hashes) != len(hashes_items_to_upload):
            problem_items = [(hsh, item, resp['errors'])
                             for (hsh, item), resp in zip(hashes_items_to_upload, resp_list) if resp.get('errors')]
            logger.warn('Not all images were uploaded within request.', extra={
                'total_cnt': len(hashes_items_to_upload), 'ok_cnt': len(remote_hashes), 'items': problem_items})
        return remote_hashes

    def _upload_data_bulk(self, func_item_to_byte_stream, items_hashes, retry_cnt=3, progress_cb=None):
        """
        Upload images (binary data) to server. Works with already existing or duplicating images.
        :param func_item_to_byte_stream: converter for "item" to byte stream
        :param items_hashes: iterable of pairs (item, hash) where "item" is a some descriptor (e.g. image file path)
         for image data, and "hash" is a hash for the image binary data
        :param retry_cnt: int, number of retries to send the whole set of items
        :param progress_cb: callback to account progress (in number of items)
        """
        hash_to_items = {i_hash: item for item, i_hash in items_hashes}

        unique_hashes = set(hash_to_items.keys())
        remote_hashes = set(self.check_existing_hashes(list(unique_hashes)))  # existing -- from server
        if progress_cb:
            progress_cb(len(remote_hashes))
        pending_hashes = unique_hashes - remote_hashes

        # @TODO: some correlation with sly.io.network_exceptions. Should we perform retries here?
        for retry_idx in range(retry_cnt):
            # single attempt to upload all data which is not uploaded yet

            for hashes in batched(list(pending_hashes)):
                pending_hashes_items = [(h, hash_to_items[h]) for h in hashes]
                hashes_rcv = self._upload_uniq_images_single_req(func_item_to_byte_stream, pending_hashes_items)
                pending_hashes -= set(hashes_rcv)
                if set(hashes_rcv) - set(hashes):
                    logger.warn('Hash inconsistency in images bulk upload.',
                                extra={'sent': hashes, 'received': hashes_rcv})
                if progress_cb:
                    progress_cb(len(hashes_rcv))

            if not pending_hashes:
                return

            logger.warn('Unable to upload images (data).', extra={
                'retry_idx': retry_idx,
                'items': [(h, hash_to_items[h]) for h in pending_hashes]
            })
            # now retry it for the case if it is a shadow server/connection error

        raise RuntimeError("Unable to upload images (data). "
                           "Please check if images are in supported format and if ones aren't corrupted.")

    def upload_path(self, dataset_id, name, path, meta=None):
        '''
        Upload image with given name from given path to dataset
        :param dataset_id: int
        :param name: str
        :param path: str
        :param meta:
        :return: list of images
        '''
        metas = None if meta is None else [meta]
        return self.upload_paths(dataset_id, [name], [path], metas=metas)[0]

    def upload_paths(self, dataset_id, names, paths, progress_cb=None, metas=None):
        '''
        Upload images with given names from given pathes to dataset
        :param dataset_id: int
        :param names: list of str (if lengh of names list != lengh of paths list raise error)
        :param paths: list of str
        :param progress_cb:
        :param metas: list of dicts
        :return: list of images
        '''
        def path_to_bytes_stream(path):
            return open(path, 'rb')

        hashes = [get_file_hash(x) for x in paths]

        self._upload_data_bulk(path_to_bytes_stream, zip(paths, hashes), progress_cb=progress_cb)
        return self.upload_hashes(dataset_id, names, hashes, metas=metas)

    def upload_np(self, dataset_id, name, img, meta=None):
        '''
        Upload given image in numpy with given name to dataset
        :param dataset_id: int
        :param name: str
        :param img: image in RGB format(numpy matrix)
        :param meta:
        :return: list of images
        '''
        metas = None if meta is None else [meta]
        return self.upload_nps(dataset_id, [name], [img], metas=metas)[0]

    def upload_nps(self, dataset_id, names, imgs, progress_cb=None, metas=None):
        '''
        Upload given images in numpy with given names to dataset
        :param dataset_id: int
        :param names: list of str (if lengh of names list != lengh of imgs list raise error)
        :param imgs: list of numpy matrix
        :param progress_cb:
        :param metas:
        :return: list of images
        '''
        def img_to_bytes_stream(item):
            img, name = item[0], item[1]
            img_bytes = sly_image.write_bytes(img, get_file_ext(name))
            return io.BytesIO(img_bytes)

        def img_to_hash(item):
            img, name = item[0], item[1]
            return sly_image.get_hash(img, get_file_ext(name))

        img_name_list = list(zip(imgs, names))
        hashes = [img_to_hash(x) for x in img_name_list]

        self._upload_data_bulk(img_to_bytes_stream, zip(img_name_list, hashes), progress_cb=progress_cb)
        return self.upload_hashes(dataset_id, names, hashes, metas=metas)

    def upload_link(self, dataset_id, name, link, meta=None):
        '''
        Upload image from given link with given name to dataset
        :param dataset_id: int
        :param name: str
        :param link: str
        :param meta:
        :return: list of images
        '''
        metas = None if meta is None else [meta]
        return self.upload_links(dataset_id, [name], [link], metas=metas)[0]

    def upload_links(self, dataset_id, names, links, progress_cb=None,  metas=None):
        '''
        Upload images from given links with given names to dataset
        :param dataset_id: int
        :param names: list of str (if lengh of names list != lengh of links list raise error)
        :param links: list of str
        :param progress_cb:
        :param metas:
        :return: list of images
        '''
        return self._upload_bulk_add(lambda item: (ApiField.LINK, item), dataset_id, names, links, progress_cb, metas=metas)

    def upload_hash(self, dataset_id, name, hash, meta=None):
        '''
        Upload image from given hash with given name to dataset
        :param dataset_id: int
        :param name: str
        :param hash: str
        :param meta:
        :return: list of images
        '''
        metas = None if meta is None else [meta]
        return self.upload_hashes(dataset_id, [name], [hash], metas=metas)[0]

    def upload_hashes(self, dataset_id, names, hashes, progress_cb=None, metas=None):
        '''
        Upload images from given hashes with given names to dataset
        :param dataset_id: int
        :param names: list of str (if lengh of names list != lengh of hashes list raise error)
        :param hashes: list of str
        :param progress_cb:
        :param metas:
        :return: list of images
        '''
        return self._upload_bulk_add(lambda item: (ApiField.HASH, item), dataset_id, names, hashes, progress_cb, metas=metas)

    def upload_id(self, dataset_id, name, id, meta=None):
        '''
        Upload image from given id with given name to dataset
        :param dataset_id: int
        :param name: str
        :param id: int
        :param meta:
        :return: list of images
        '''
        metas = None if meta is None else [meta]
        return self.upload_ids(dataset_id, [name], [id], metas=metas)[0]

    def upload_ids(self, dataset_id, names, ids, progress_cb=None, metas=None):
        if metas is None:
            metas = [{}] * len(names)

        infos = self.get_info_by_id_batch(ids)

        # prev implementation
        #hashes = [info.hash for info in infos]
        #return self.upload_hashes(dataset_id, names, hashes, progress_cb, metas=metas)

        links, links_names, links_order, links_metas = [], [], [], []
        hashes, hashes_names, hashes_order, hashes_metas = [], [], [], []
        for idx, (name, info, meta) in enumerate(zip(names, infos, metas)):
            if info.link is not None:
                links.append(info.link)
                links_names.append(name)
                links_order.append(idx)
                links_metas.append(meta)
            else:
                hashes.append(info.hash)
                hashes_names.append(name)
                hashes_order.append(idx)
                hashes_metas.append(meta)

        result = [None] * len(names)
        if len(links) > 0:
            res_infos_links = self.upload_links(dataset_id, links_names, links, progress_cb, metas=links_metas)
            for info, pos in zip(res_infos_links, links_order):
                result[pos] = info

        if len(hashes) > 0:
            res_infos_hashes = self.upload_hashes(dataset_id, hashes_names, hashes, progress_cb, metas=hashes_metas)
            for info, pos in zip(res_infos_hashes, hashes_order):
                result[pos] = info

        return result

    def _upload_bulk_add(self, func_item_to_kv, dataset_id, names, items, progress_cb=None, metas=None):
        results = []

        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError("Can not match \"names\" and \"items\" lists, len(names) != len(items)")

        if metas is None:
            metas = [{}] * len(names)
        else:
            if len(names) != len(metas):
                raise RuntimeError("Can not match \"names\" and \"metas\" len(names) != len(metas)")

        for batch in batched(list(zip(names, items, metas))):
            images = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                #@TODO: 'title' -> ApiField.NAME
                image_data = {'title': name, item_tuple[0]: item_tuple[1]}
                if len(meta) != 0 and type(meta) == dict:
                    image_data[ApiField.META] = meta
                images.append(image_data)

            response = self._api.post('images.bulk.add', {ApiField.DATASET_ID: dataset_id, ApiField.IMAGES: images})
            if progress_cb is not None:
                progress_cb(len(images))

            for info_json in response.json():
                info_json_copy = info_json.copy()
                info_json_copy[ApiField.EXT] = info_json[ApiField.MIME].split('/')[1]
                #results.append(self.InfoType(*[info_json_copy[field_name] for field_name in self.info_sequence()]))
                results.append(self._convert_json_info(info_json_copy))

        #name_to_res = {img_info.name: img_info for img_info in results}
        #ordered_results = [name_to_res[name] for name in names]

        return results #ordered_results

    #@TODO: reimplement
    def _convert_json_info(self, info: dict, skip_missing=True):
        if info is None:
            return None
        temp_ext = None
        field_values = []
        for field_name in self.info_sequence():
            if field_name == ApiField.EXT:
                continue
            if skip_missing is True:
                val = info.get(field_name, None)
            else:
                val = info[field_name]
            field_values.append(val)
            if field_name == ApiField.MIME:
                temp_ext = val.split('/')[1]
                field_values.append(temp_ext)
        for idx, field_name in enumerate(self.info_sequence()):
            if field_name == ApiField.NAME:
                cur_ext = get_file_ext(field_values[idx]).replace(".", "").lower()
                if not cur_ext:
                    field_values[idx] = "{}.{}".format(field_values[idx], temp_ext)
                    break
                if temp_ext == 'jpeg' and cur_ext in ['jpg', 'jpeg', 'mpo']:
                    break
                if temp_ext != cur_ext:
                    field_values[idx] = "{}.{}".format(field_values[idx], temp_ext)
                break
        return self.InfoType(*field_values)

    def _remove_batch_api_method_name(self):
        return 'images.bulk.remove'

    def _remove_batch_field_name(self):
        return ApiField.IMAGE_IDS

    def copy_batch(self, dst_dataset_id, ids, change_name_if_conflict=False, with_annotations=False):
        '''
        Copy images with given ids in destination dataset
        :param dst_dataset_id: int
        :param ids: list of integers (raise error if type of ids not listand if images ids have to be from the same dataset)
        :param change_name_if_conflict: bool (if False and images with the same names already exist in destination dataset raise error)
        :param with_annotations: bool (if True - annotations will be copy to dataset too)
        :return: list of images
        '''
        if type(ids) is not list:
            raise RuntimeError("ids parameter has type {!r}. but has to be of type {!r}".format(type(ids), list))

        if len(ids) == 0:
            return

        existing_images = self.get_list(dst_dataset_id)
        existing_names = {image.name for image in existing_images}

        ids_info = self.get_info_by_id_batch(ids)
        temp_ds_ids = {info.dataset_id for info in ids_info}
        if len(temp_ds_ids) > 1:
            raise RuntimeError("Images ids have to be from the same dataset")

        if change_name_if_conflict:
            new_names = [generate_free_name(existing_names, info.name, with_ext=True) for info in ids_info]
        else:
            new_names = [info.name for info in ids_info]
            names_intersection = existing_names.intersection(set(new_names))
            if len(names_intersection) != 0:
                raise RuntimeError('Images with the same names already exist in destination dataset. '
                                   'Please, use argument \"change_name_if_conflict=True\" to automatically resolve '
                                   'names intersection')

        new_images = self.upload_ids(dst_dataset_id, new_names, ids)
        new_ids = [new_image.id for new_image in new_images]

        if with_annotations:
            src_project_id = self._api.dataset.get_info_by_id(ids_info[0].dataset_id).project_id
            dst_project_id = self._api.dataset.get_info_by_id(dst_dataset_id).project_id
            self._api.project.merge_metas(src_project_id, dst_project_id)
            self._api.annotation.copy_batch(ids, new_ids)

        return new_images

    def move_batch(self, dst_dataset_id, ids, change_name_if_conflict=False, with_annotations=False):
        '''
        Move images with given ids in destination dataset
        :param dst_dataset_id: int
        :param ids: list of integers (raise error if type of ids not listand if images ids have to be from the same dataset)
        :param change_name_if_conflict: bool (if False and images with the same names already exist in destination dataset raise error)
        :param with_annotations: bool (if True - annotations will be copy to dataset too)
        :return: list of images
        '''
        new_images = self.copy_batch(dst_dataset_id, ids, change_name_if_conflict, with_annotations)
        self.remove_batch(ids)
        return new_images

    def copy(self, dst_dataset_id, id, change_name_if_conflict=False, with_annotations=False):
        '''
        Copy image with given id in destination dataset
        :param dst_dataset_id: int
        :param id: int
        :param change_name_if_conflict: bool (if False and images with the same names already exist in destination dataset raise error)
        :param with_annotations: bool
        :return: image
        '''
        return self.copy_batch(dst_dataset_id, [id], change_name_if_conflict, with_annotations)[0]

    def move(self, dst_dataset_id, id, change_name_if_conflict=False, with_annotations=False):
        '''
        Move image with given id in destination dataset
        :param dst_dataset_id: int
        :param id: int
        :param change_name_if_conflict: bool (if False and images with the same names already exist in destination dataset raise error)
        :param with_annotations: bool
        :return: image
        '''
        return self.move_batch(dst_dataset_id, [id], change_name_if_conflict, with_annotations)[0]

    def url(self, team_id, workspace_id, project_id, dataset_id, image_id):
        '''
        :param team_id: int
        :param workspace_id: int
        :param project_id: int
        :param dataset_id: int
        :param image_id: int
        :return: url of given image id
        '''
        result = urllib.parse.urljoin(self._api.server_address,
                                      'app/images/{}/{}/{}/{}#image-{}'.format(team_id,
                                                                               workspace_id,
                                                                               project_id,
                                                                               dataset_id,
                                                                               image_id)
                                      )

        return result

    def _download_batch_by_hashes(self, hashes):
        for batch_hashes in batched(hashes):
            response = self._api.post(
                'images.bulk.download-by-hash', {ApiField.HASHES: batch_hashes})
            decoder = MultipartDecoder.from_response(response)
            for part in decoder.parts:
                content_utf8 = part.headers[b'Content-Disposition'].decode('utf-8')
                # Find name="1245" preceded by a whitespace, semicolon or beginning of line.
                # The regex has 2 capture group: one for the prefix and one for the actual name value.
                h = content_utf8.replace("form-data; name=\"", "")[:-1]
                yield h, part

    def download_paths_by_hashes(self, hashes, paths, progress_cb=None):
        if len(hashes) == 0:
            return
        if len(hashes) != len(paths):
            raise RuntimeError("Can not match \"hashes\" and \"paths\" lists, len(hashes) != len(paths)")

        h_to_path = {h: path for h, path in zip(hashes, paths)}
        for h, resp_part in self._download_batch_by_hashes(list(set(hashes))):
            ensure_base_path(h_to_path[h])
            with open(h_to_path[h], 'wb') as w:
                w.write(resp_part.content)
            if progress_cb is not None:
                progress_cb(1)
                
    def get_project_id(self, image_id):
        dataset_id = self.get_info_by_id(image_id).dataset_id
        project_id = self._api.dataset.get_info_by_id(dataset_id).project_id
        return project_id

        
    @staticmethod
    def _get_free_name(exist_check_fn, name):
        res_title = name
        suffix = 1

        name_without_ext = get_file_name(name)
        ext = get_file_ext(name)

        while exist_check_fn(res_title):
            res_title = '{}_{:03d}{}'.format(name_without_ext, suffix, ext)
            suffix += 1
        return res_title

    def storage_url(self, path_original):
        result = urllib.parse.urljoin(self._api.server_address, '{}'.format(path_original))
        return result

    def preview_url(self, url, width=None, height=None, quality=70):
        #@TODO: if both width and height are defined, and they are not proportioned to original image resolution,
        # then images will be croped from center
        if width is None:
            width = ""
        if height is None:
            height = ""
        return url.replace("/image-converter", "/previews/{}x{},jpeg,q{}/image-converter".format(width, height, quality))

    def update_meta(self, id, meta):
        if type(meta) is not dict:
            raise TypeError('Meta must be dict, not {}'.format(type(meta)))
        response = self._api.post('images.editInfo', {ApiField.ID: id, ApiField.META: meta})
        return response.json()

    def add_tag(self, image_id, tag_id, value=None):
        # data = {ApiField.TAG_ID: tag_id, ApiField.IMAGE_ID: image_id}
        # if value is not None:
        #     data[ApiField.VALUE] = value
        # resp = self._api.post('image-tags.add-to-image', data)
        # return resp.json()
        self.add_tag_batch([image_id], tag_id, value)

    def add_tag_batch(self, image_ids, tag_id, value=None):
        data = {ApiField.TAG_ID: tag_id, ApiField.IDS: image_ids}
        if value is not None:
            data[ApiField.VALUE] = value
        resp = self._api.post('image-tags.bulk.add-to-image', data)
        return resp.json()
